# -*- coding: utf-8 -*-
"""
Routines for practical session on image reconstruction and inverse problems.

To start your session, just do:

    from tp_utils import *          # utilities for the practical session

Work in progress ;-)
"""
# Insure compatibility.
from __future__ import absolute_import, division, print_function, unicode_literals

# Import modules using the same conventions as SciPy:
import numpy as np              # to deal with numerical arrays
import matplotlib as mpl        # for graphics
import matplotlib.pyplot as plt # for plotting
from astropy.io import fits

# Import some NumPy routines.
from numpy import conj, real, imag

# Other imports.
import math

# Import some routines for graphics.
figure = plt.figure
clf = plt.clf
cla = plt.cla
fma = plt.clf
imshow = plt.imshow
close = plt.close
show = plt.show
get_cmap = plt.get_cmap
colorbar = plt.colorbar

# Turn interactive graphics on.
plt.ion()

# Aliases/shortcuts for FFT routines.
fftshift = np.fft.fftshift
ifftshift = np.fft.ifftshift
fft = np.fft.fftn
def ifft(x): return real(np.fft.ifftn(x))

def abs2(x):
    """Returns the squared absolute value of its agument."""
    if np.iscomplexobj(x):
        x_re = x.real
        x_im = x.imag
        return x_re*x_re + x_im*x_im
    else:
        return x*x

def normalize(psf):
    s = sum(psf.ravel())
    print("sum(PSF) = {0}".format(s))
    if s == 1.0: return psf
    if s > 0.0: return (1.0/s)*psf
    print('WARNING sum(PSF) = {0}'.format(s))
    return psf

def fits_read(filename):
    """Read a FITS image from a file."""
    return fits.open(filename)[0].data

def fits_write(filename, data, comment=None, history=None):
    """Write an image into a FITS file."""
    hdu = fits.PrimaryHDU(data)
    if comment is not None:
        hdu['COMMENT'] = comment
    if history is not None:
        hdu['HISTORY'] = history
    hdu.writeto(filename)

def conjgrad(A, b, x=None, maxiter=50, gatol=0.0, grtol=1e-5, quiet=False, viewer=None):
    """Conjugate gradient routine.

    Solve the linear problem `A.x = b` by means the conjugate gradient
    method.

    Argument `A` is a function to compute the product of the "lef-hand-side
    (LHS) matrix" of the problem with a vector given in argument.

    Argument `b` is the "right-hand-side (RHS) vector".

    Optional argument `x` is an initial solution to start with.  By
    default, the initial solution is zero everywhere.

    Optional arguments `gatol` and `grtol` (0.0 and 1E-5 respectively by
    default) are the absolute and relative gradient tolerance for
    convergence.  The conjugate gradient iterations are performed until the
    squared Euclidean norm of the residual is less or equal the largest
    between `gatol` and `grtol` times the squared Euclidean norm of the
    initial residuals.

    Optional argument `maxiter` (50 by default) is the maximum number of
    iterations to perform until returning the result.  A warning message is
    printed if the number of iterations is exceeded (unless `quiet` is
    true).
    """

    def inner(x,y):
        """Compute inner product of X and Y regardless their shapes
        (their number of elements must however match)."""
        return np.inner(x.ravel(),y.ravel())

    # Compute initial residuals.
    if x is None:
        r = b
        x = np.zeros(b.shape, dtype=b.dtype)
    else:
        r = b - A(x)
    rho = inner(r,r)
    epsilon = max(gatol, grtol*math.sqrt(rho))

    # Conjugate gradient iterations.
    beta = 0.0
    k = 0
    while True:
        if viewer is not None:
            viewer(x, k, math.sqrt(rho))
        if math.sqrt(rho) <= epsilon:
            break
        k += 1
        if k > maxiter:
            if not quiet:
                print("WARNING: too many iterations ({0})".format(maxiter))
            break

        # Next search direction.
        if beta == 0.0:
            p = r
        else:
            p = r + beta*p

        # Make optimal step along search direction.
        q = A(p)
        gamma = inner(p, q)
        if gamma <= 0.0:
            raise ValueError("Operator A is not positive definite")
        alpha = rho/gamma
        x += alpha*p
        r -= alpha*q
        rho_prev, rho = rho, inner(r,r)
        beta = rho/rho_prev
    return x

def zeropad(x, shape):
    """Zero-pad array X to given shape.

    The contents of X is approximately centered in the result."""
    rank = x.ndim
    if len(shape) != rank:
        raise ValueError("bad number of dimensions")
    diff = np.array(shape) - np.array(x.shape)
    if diff.min() < 0:
        raise ValueError("output dimensions must be larger or equal input dimensions")
    offset = diff//2
    z = np.zeros(shape, dtype=x.dtype)
    if rank == 1:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        z[i0:n0] = x
    elif rank == 2:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        i1 = offset[1]; n1 = i1 + x.shape[1]
        z[i0:n0,i1:n1] = x
    elif rank == 3:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        i1 = offset[1]; n1 = i1 + x.shape[1]
        i2 = offset[2]; n2 = i2 + x.shape[2]
        z[i0:n0,i1:n1,i2:n2] = x
    elif rank == 4:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        i1 = offset[1]; n1 = i1 + x.shape[1]
        i2 = offset[2]; n2 = i2 + x.shape[2]
        i3 = offset[3]; n3 = i3 + x.shape[3]
        z[i0:n0,i1:n1,i2:n2,i3:n3] = x
    elif rank == 5:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        i1 = offset[1]; n1 = i1 + x.shape[1]
        i2 = offset[2]; n2 = i2 + x.shape[2]
        i3 = offset[3]; n3 = i3 + x.shape[3]
        i4 = offset[4]; n4 = i4 + x.shape[4]
        z[i0:n0,i1:n1,i2:n2,i3:n3,i4:n4] = x
    elif rank == 6:
        i0 = offset[0]; n0 = i0 + x.shape[0]
        i1 = offset[1]; n1 = i1 + x.shape[1]
        i2 = offset[2]; n2 = i2 + x.shape[2]
        i3 = offset[3]; n3 = i3 + x.shape[3]
        i4 = offset[4]; n4 = i4 + x.shape[4]
        i5 = offset[5]; n5 = i5 + x.shape[5]
        z[i0:n0,i1:n1,i2:n2,i3:n3,i4:n4,i5:n5] = x
    else:
        raise ValueError("too many dimensions")
    return z

def DtD(x):
    """Returns the result of D'.D.x where D is a (multi-dimensional)
    finite difference operator and D' is its transpose."""

    dims = x.shape
    r = np.zeros(dims, dtype=x.dtype) # to store the result
    rank = x.ndim # number of dimensions
    if rank == 0: return r
    if dims[0] >= 2:
        dx = x[1:-1,...] - x[0:-2,...]
        r[1:-1,...] += dx
        r[0:-2,...] -= dx
    if rank == 1: return r
    if dims[1] >= 2:
        dx = x[:,1:-1,...] - x[:,0:-2,...]
        r[:,1:-1,...] += dx
        r[:,0:-2,...] -= dx
    if rank == 2: return r
    if dims[2] >= 2:
        dx = x[:,:,1:-1,...] - x[:,:,0:-2,...]
        r[:,:,1:-1,...] += dx
        r[:,:,0:-2,...] -= dx
    if rank == 3: return r
    if dims[3] >= 2:
        dx = x[:,:,:,1:-1,...] - x[:,:,:,0:-2,...]
        r[:,:,:,1:-1,...] += dx
        r[:,:,:,0:-2,...] -= dx
    if rank == 4: return r
    if dims[4] >= 2:
        dx = x[:,:,:,:,1:-1,...] - x[:,:,:,:,0:-2,...]
        r[:,:,:,:,1:-1,...] += dx
        r[:,:,:,:,0:-2,...] -= dx
    if rank == 5: return r
    raise ValueError("too many dimensions")


if __name__ == "__main__":
    print("testing TiPi!")

